#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>
#include <zlib.h>

#include "pktlib.h"
#include "pktbuf.h"
#include "arplib.h"
#include "nat.h"

static char *make_srcmacaddr(in_addr_t ipaddr)
{
  static char macaddr[ETHER_ADDR_LEN];
  ipaddr = htonl(ipaddr);
  macaddr[0] = 0x00;
  macaddr[1] = 0x11;
  memcpy(&macaddr[2], &ipaddr, sizeof(in_addr_t));
  return macaddr;
}

static in_addr_t proc_ip(pktbuf_t pktbuf, in_addr_t my_ipaddr)
{
  char *p;
  struct ip *iphdr;
  struct icmp *icmphdr;
  int hdrsize, paysize;

  p = pktbuf_get_header(pktbuf);

  if (pktbuf_get_size(pktbuf) < sizeof(*iphdr))
    return 0;

  iphdr = (struct ip *)p;
  hdrsize = iphdr->ip_hl << 2;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  if (iphdr->ip_p != IPPROTO_ICMP)
    return 0;
  if (ntohl(iphdr->ip_dst.s_addr) != my_ipaddr)
    return 0;

  if (pktbuf_get_size(pktbuf) < hdrsize + paysize)
    return 0;

  if (paysize < sizeof(*icmphdr))
    return 0;

  icmphdr = (struct icmp *)(p + hdrsize);
  if (icmphdr->icmp_type != ICMP_ECHO)
    return 0;

  fprintf(stderr, "RECV ICMP Echo       %-15s", inet_ntoa(iphdr->ip_src));
  fprintf(stderr, " -> %-15s\n", inet_ntoa(iphdr->ip_dst));

  icmphdr->icmp_type = ICMP_ECHOREPLY;
  icmphdr->icmp_cksum = 0;

  iphdr->ip_dst.s_addr = iphdr->ip_src.s_addr;
  iphdr->ip_src.s_addr = htonl(my_ipaddr);
  iphdr->ip_sum = 0;

  fprintf(stderr, "SEND ICMP Echo Reply %-15s", inet_ntoa(iphdr->ip_src));
  fprintf(stderr, " -> %-15s\n", inet_ntoa(iphdr->ip_dst));

  return ntohl(iphdr->ip_dst.s_addr);
}

static in_addr_t proc_ttl(pktbuf_t pktbuf)
{
  char *p;
  struct ip *iphdr;

  p = pktbuf_get_header(pktbuf);

  if (pktbuf_get_size(pktbuf) < sizeof(*iphdr))
    return 0;

  iphdr = (struct ip *)p;
  if (iphdr->ip_ttl <= 1)
    return 0;

  iphdr->ip_ttl--;

  return ntohl(iphdr->ip_dst.s_addr);
}

#define SPECIAL_CRYPT_PROTO_FLAG 0x80

static pktbuf_t encrypt_ip(pktbuf_t pktbuf, int key)
{
  char *p;
  struct ip *iphdr;
  int hdrsize, paysize;
  int i;

  p = pktbuf_get_header(pktbuf);

  if (pktbuf_get_size(pktbuf) < sizeof(*iphdr))
    return pktbuf;

  iphdr = (struct ip *)p;
  hdrsize = iphdr->ip_hl << 2;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  if (pktbuf_get_size(pktbuf) < hdrsize + paysize)
    return pktbuf;

  for (i = 0; i < paysize; i++) {
    (p + hdrsize)[i] ^= key;
  }

  iphdr->ip_p |= SPECIAL_CRYPT_PROTO_FLAG;

  return pktbuf;
}

static pktbuf_t decrypt_ip(pktbuf_t pktbuf, int key)
{
  char *p;
  struct ip *iphdr;
  int hdrsize, paysize;
  int i;

  p = pktbuf_get_header(pktbuf);

  if (pktbuf_get_size(pktbuf) < sizeof(*iphdr))
    return pktbuf;

  iphdr = (struct ip *)p;
  hdrsize = iphdr->ip_hl << 2;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  if (!(iphdr->ip_p & SPECIAL_CRYPT_PROTO_FLAG))
    return pktbuf;

  if (pktbuf_get_size(pktbuf) < hdrsize + paysize)
    return pktbuf;

  for (i = 0; i < paysize; i++) {
    (p + hdrsize)[i] ^= key;
  }

  iphdr->ip_p &= ~SPECIAL_CRYPT_PROTO_FLAG;

  return pktbuf;
}

#define SPECIAL_COMPRESS_PROTO_FLAG 0x40

static pktbuf_t compress_ip(pktbuf_t pktbuf)
{
  char *p;
  struct ip *iphdr;
  int hdrsize, paysize;
  int offset = 0, woffset = 0, size, s, ret, flush;
  z_stream z;
  char ibuf[32*1024];
  char obuf[32*1024];
  char wbuf[64*1024];

  p = pktbuf_get_header(pktbuf);

  if (pktbuf_get_size(pktbuf) < sizeof(*iphdr))
    return pktbuf;

  iphdr = (struct ip *)p;
  hdrsize = iphdr->ip_hl << 2;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  if (pktbuf_get_size(pktbuf) < hdrsize + paysize)
    return pktbuf;

  size = paysize;

  z.zalloc = Z_NULL;
  z.zfree = Z_NULL;
  z.opaque = Z_NULL;

  ret = deflateInit(&z, Z_DEFAULT_COMPRESSION);
  if (ret != Z_OK)
    return pktbuf;

  z.next_in = Z_NULL;
  z.avail_in = 0;
  z.next_out = (Bytef *)obuf;
  z.avail_out = sizeof(obuf);

  flush = Z_NO_FLUSH;

  do {
    if (z.avail_in == 0) {
      s = sizeof(ibuf);
      if (s > size) s = size;
      memcpy(ibuf, p + hdrsize + offset, s);
      offset += s;
      size -= s;
      z.next_in = (Bytef *)ibuf;
      z.avail_in = s;
      if (size == 0)
	flush = Z_FINISH;
    }

    ret = deflate(&z, flush);
    if ((ret != Z_OK) && (ret != Z_STREAM_END))
      return pktbuf;

    if ((z.avail_out == 0) || (ret == Z_STREAM_END)) {
      s = sizeof(obuf) - z.avail_out;
      memcpy(wbuf + woffset, obuf, s);
      woffset += s;
      z.next_out = (Bytef *)obuf;
      z.avail_out = sizeof(obuf);
    }
  } while (ret != Z_STREAM_END);

  if (deflateEnd(&z) != Z_OK)
    return pktbuf;

  paysize = woffset;

  memcpy(p + hdrsize, wbuf, paysize);

  pktbuf_set_size(pktbuf, hdrsize + paysize);

  iphdr->ip_p |= SPECIAL_COMPRESS_PROTO_FLAG;
  iphdr->ip_len = htons(hdrsize + paysize);

  return pktbuf;
}

static pktbuf_t decompress_ip(pktbuf_t pktbuf)
{
  char *p;
  struct ip *iphdr;
  int hdrsize, paysize;
  int offset = 0, woffset = 0, size, s, ret;
  z_stream z;
  char ibuf[32*1024];
  char obuf[32*1024];
  char wbuf[64*1024];

  p = pktbuf_get_header(pktbuf);

  if (pktbuf_get_size(pktbuf) < sizeof(*iphdr))
    return pktbuf;

  iphdr = (struct ip *)p;
  hdrsize = iphdr->ip_hl << 2;
  paysize = ntohs(iphdr->ip_len) - hdrsize;

  if (!(iphdr->ip_p & SPECIAL_COMPRESS_PROTO_FLAG))
    return pktbuf;

  if (pktbuf_get_size(pktbuf) < hdrsize + paysize)
    return pktbuf;

  size = paysize;

  z.zalloc = Z_NULL;
  z.zfree = Z_NULL;
  z.opaque = Z_NULL;

  z.next_in = Z_NULL;
  z.avail_in = 0;
  ret = inflateInit(&z);
  if (ret != Z_OK)
    return pktbuf;

  z.next_out = (Bytef *)obuf;
  z.avail_out = sizeof(obuf);

  do {
    if (z.avail_in == 0) {
      s = sizeof(ibuf);
      if (s > size) s = size;
      memcpy(ibuf, p + hdrsize + offset, s);
      offset += s;
      size -= s;
      z.next_in = (Bytef *)ibuf;
      z.avail_in = s;
    }

    ret = inflate(&z, Z_NO_FLUSH);
    if ((ret != Z_OK) && (ret != Z_STREAM_END))
      return pktbuf;

    if ((z.avail_out == 0) || (ret == Z_STREAM_END)) {
      s = sizeof(obuf) - z.avail_out;
      memcpy(wbuf + woffset, obuf, s);
      woffset += s;
      z.next_out = (Bytef *)obuf;
      z.avail_out = sizeof(obuf);
    }
  } while (ret != Z_STREAM_END);

  if (inflateEnd(&z) != Z_OK)
    return pktbuf;

  paysize = woffset;

  memcpy(p + hdrsize, wbuf, paysize);

  pktbuf_set_size(pktbuf, hdrsize + paysize);

  iphdr->ip_p &= ~SPECIAL_COMPRESS_PROTO_FLAG;
  iphdr->ip_len = htons(hdrsize + paysize);

  return pktbuf;
}

struct ip_iface {
  pktif_t pktif;
  char macaddr[ETHER_ADDR_LEN];
  in_addr_t ipaddr;
  in_addr_t ipmask;
  arptbl_t arptbl;
  natif_t natif;
  int compress;
  int key;
};

struct ip_route {
  in_addr_t ipaddr;
  in_addr_t ipmask;
  in_addr_t gateway;
  struct ip_iface *ifp;
};

int main(int argc, char *argv[])
{
  pktif_t pktif;
  pktbuf_t pktbuf;
  int ifnum = 0, rtnum = 0;
  in_addr_t dstaddr;
  struct in_addr in;
  struct ether_header *ehdr;
  struct ip *iphdr;
  struct ip_iface iflist[16], *ifp = NULL;
  struct ip_route rtlist[16], *rtp = rtlist, *found;

  argc--; argv++;
  /* 󥿡ե */
  for (ifp = iflist; argc > 5; argc -= 6, argv += 6) {
    if (!strcmp(argv[0], "--"))
      break;
    /* 󥿡ե򥪡ץ󤹤(ȯѥåȤϼʤ) */
    ifp->pktif = pktif_open(argv[0], PKTIF_OPEN_FLAG_RECV_NOTSENT, 0);
    ifp->ipaddr = ntohl(inet_addr(argv[1])); /* IPɥ쥹 */
    ifp->ipmask = ntohl(inet_addr(argv[2])); /* ͥåȥޥ */
    memcpy(ifp->macaddr, make_srcmacaddr(ifp->ipaddr), ETHER_ADDR_LEN);
    ifp->arptbl = arplib_table_create(ifp->ipaddr, ifp->macaddr);
    ifp->natif = NULL;
    ifp->compress = 0;
    ifp->key = 0;
    if (!strcmp(argv[3], "nat"))
      ifp->natif = natif_create(1024, ifp->ipaddr);
    if (!strcmp(argv[4], "compress"))
      ifp->compress = 1;
    if (isdigit(argv[5][0]))
      ifp->key = atoi(argv[5]);

    /* ͥƥåɡ롼Ȥ */
    rtp->ipaddr = ifp->ipaddr & ifp->ipmask;
    rtp->ipmask = ifp->ipmask;
    rtp->gateway = 0;
    rtp->ifp = ifp;
    rtnum++;
    rtp++;

    ifnum++;
    ifp++;
  }

  argc--; argv++;
  /* ϩ */
  for (; argc > 2; argc -= 3, argv += 3) {
    if (!strcmp(argv[0], "--"))
      break;
    rtp->ipaddr  = ntohl(inet_addr(argv[0])); /* IPɥ쥹 */
    rtp->ipmask  = ntohl(inet_addr(argv[1])); /* ͥåȥޥ */
    rtp->gateway = ntohl(inet_addr(argv[2])); /* ž */
    rtp->ifp = NULL;
    rtnum++;
    rtp++;
  }

  argc--; argv++;
  /* ŪNATȥ */
  for (; argc > 4; argc -= 5, argv += 5) {
    for (ifp = iflist; ifp < &iflist[ifnum]; ifp++) {
      if (!strcmp(pktif_get_name(ifp->pktif), argv[0]))
	break;
    }
    if (ifp != &iflist[ifnum]) {
      switch (argv[1][0]) {
      case 't':
	natif_add_entry_tcp(ifp->natif, atoi(argv[2]),
			    ntohl(inet_addr(argv[3])), atoi(argv[4]));
	break;
      case 'u':
	natif_add_entry_udp(ifp->natif, atoi(argv[2]),
			    ntohl(inet_addr(argv[3])), atoi(argv[4]));
	break;
      default:
	break;
      }
    }
  }

  fprintf(stderr, "Interfaces:\n");
  for (ifp = iflist; ifp < &iflist[ifnum]; ifp++) {
    fprintf(stderr, "\t%s", pktif_get_name(ifp->pktif));
    in.s_addr = htonl(ifp->ipaddr);
    fprintf(stderr, " %s", inet_ntoa(in));
    in.s_addr = htonl(ifp->ipmask);
    fprintf(stderr, "/%s", inet_ntoa(in));
    fprintf(stderr, " (%s)", (ifp->natif) ? "nat" : "normal");
    fprintf(stderr, " (%s)", (ifp->compress) ? "compress" : "normal");
    fprintf(stderr, " (0x%02x)", ifp->key);
    fprintf(stderr, "\n");
  }

  fprintf(stderr, "Routes:\n");
  for (rtp = rtlist; rtp < &rtlist[rtnum]; rtp++) {
    in.s_addr = htonl(rtp->ipaddr);
    fprintf(stderr, "\t%s", inet_ntoa(in));
    in.s_addr = htonl(rtp->ipmask);
    fprintf(stderr, "/%s", inet_ntoa(in));
    if (rtp->ifp) {
      fprintf(stderr, " -> %s\n", pktif_get_name(rtp->ifp->pktif));
    } else {
      in.s_addr = htonl(rtp->gateway);
      fprintf(stderr, " -> %s\n", inet_ntoa(in));
    }
  }

  while (1) {
    pktif = pktlib_iflist_select(-1); /* ʣ󥿡եԤ */
    if (pktif == NULL) continue;

    pktbuf = pktbuf_recv(pktif, 0); /* ѥåȤ */
    if (pktbuf == NULL) continue;

    /* 󥿡ե򸡺 */
    for (ifp = iflist; ifp < &iflist[ifnum]; ifp++) {
      if (pktif == ifp->pktif)
	break;
    }
    if (ifp == &iflist[ifnum])
      goto discard;

    ehdr = (struct ether_header *)pktbuf_get_header(pktbuf);

    /* MACɥ쥹åβݤȽǤ */
    if (!(ehdr->ether_dhost[0] & 1) &&
	memcmp(ehdr->ether_dhost, ifp->macaddr, ETHER_ADDR_LEN))
      goto discard;

    /* ARPν */
    if (ntohs(ehdr->ether_type) == ETHERTYPE_ARP) {
      pktbuf = arplib_packet_proc(ifp->arptbl, pktbuf);
      pktbuf = pktbuf_send_queue(ifp->pktif, pktbuf); /* ʣѥåȤ */
      goto discard;
    }

    if (ntohs(ehdr->ether_type) != ETHERTYPE_IP)
      goto discard;

    pktbuf_delete_header(pktbuf, ETHER_HDR_LEN);

    iphdr = (struct ip *)pktbuf_get_header(pktbuf);

    fprintf(stderr, "RECV IP Packet %s %-15s",
	    pktif_get_name(ifp->pktif), inet_ntoa(iphdr->ip_src));
    fprintf(stderr, " -> %-15s\n", inet_ntoa(iphdr->ip_dst));

    if (ifp->key)
      pktbuf = decrypt_ip(pktbuf, ifp->key);

    if (ifp->compress)
      pktbuf = decompress_ip(pktbuf);

    if (ifp->natif) { /* NATν() */
      pktbuf = natif_packet_incoming(ifp->natif, pktbuf);
      if (pktbuf == NULL)
	goto discard;
    }

    if ((ntohl(iphdr->ip_dst.s_addr) & 0xF0000000) == 0xE0000000)
      goto discard; /* IPޥ㥹Ȥ˴ */

    /* ȰΥѥåȤν */
    dstaddr = 0;
    for (ifp = iflist; ifp < &iflist[ifnum]; ifp++) {
      if ((ntohl(iphdr->ip_dst.s_addr) | ifp->ipmask) == 0xFFFFFFFF)
	goto discard; /* IP֥ɥ㥹Ȥ˴ */
      dstaddr = proc_ip(pktbuf, ifp->ipaddr); /* ICMP Echo ʤб */
      if (dstaddr)
	break;
    }
    if (dstaddr == 0) {
      dstaddr = proc_ttl(pktbuf);
      if (dstaddr == 0)
	goto discard;
    }
    iphdr = (struct ip *)pktbuf_get_header(pktbuf);
    dstaddr = ntohl(iphdr->ip_dst.s_addr);

    /* ϩ򸡺 */
    while (1) {
      found = NULL;
      for (rtp = rtlist; rtp < &rtlist[rtnum]; rtp++) {
	if ((dstaddr & rtp->ipmask) == (rtp->ipaddr & rtp->ipmask)) {
	  /* ʣηϩҥåȤˤϡ󥲥ȥޥåΥ롼Ƿ */
	  if ((found == NULL) || (rtp->ipmask > found->ipmask))
	    found = rtp;
	}
      }
      if (found == NULL) /* ϩĤʤ */
	break;
      ifp = found->ifp;
      if (ifp) /* Υ󥿡եꤷ */
	goto sendpkt;
      dstaddr = found->gateway; /* žФƷϩƸ */
    }

    goto discard;

  sendpkt:
    if (ifp->natif) { /* NATν() */
      pktbuf = natif_packet_outgoing(ifp->natif, pktbuf);
      if (pktbuf == NULL)
	goto discard;
    }

    if (ifp->compress)
      pktbuf = compress_ip(pktbuf);

    if (ifp->key)
      pktbuf = encrypt_ip(pktbuf, ifp->key);

    in.s_addr = htonl(dstaddr);
    fprintf(stderr, "SEND IP Packet %s %-15s\n",
	    pktif_get_name(ifp->pktif), inet_ntoa(in));
    pktbuf = pktbuf_checksum_correct_ip(pktbuf);
    pktbuf = arplib_packet_make(ifp->arptbl, pktbuf, ETHERTYPE_IP, dstaddr);
    pktbuf = pktbuf_send_queue(ifp->pktif, pktbuf); /* ʣѥåȤ */

  discard:
    pktbuf_destroy_queue(pktbuf);
  }

  return 0;
}
